import os

__version__ = '0.1.7'   
__root_dir__ = os.path.dirname(__file__)


from langchain.chat_models import ChatOpenAI
from chatmof.agents.agent import ChatMOF
from chatmof.config import config


__all__ = [
    "ChatMOF",
    "config",
    "ChatOpenAI",
    "__version__",
]



from langchain.chat_models import ChatOpenAI
from langchain.callbacks import StdOutCallbackHandler
from langchain.callbacks.manager import CallbackManagerForChainRun
from chatmof.config import config
from chatmof.agents.agent import ChatMOF


def main(**kwargs) -> str:
    config.update(kwargs)

    model = config['model']
    search_internet = config['search_internet']
    verbose = config['verbose']

    llm = ChatOpenAI(temperature=config['temperature'], model=model)
    callback_manager = [StdOutCallbackHandler()]
    run_manager = CallbackManagerForChainRun.get_noop_manager()

    chatmof = ChatMOF.from_llm(
        llm=llm, 
        verbose=verbose, 
        search_internet=search_internet,
    )

    print ('#' * 50 + "\n")
    print ('Welcom to ChatMOF!')
    print ("\n" + "#"*10 + ' Question ' + "#"*30)
    print ('Please enter the question below >>')
    question = input()
    
    output = chatmof.run(question, callbacks=callback_manager)

    print ('\n')
    print ("#"*10 + ' Output ' + "#" * 30)
    print (output)
    print ('\n')
    print ('Thanks for using CHATMOF!')

    return output


if __name__ == '__main__':
    main()


class CLICommand:
    """
    Run ChatMOF code

    ex) chatmof run model_name=gpt-3.5-turbo temperature=0.1 verbose=True
    """

    @staticmethod
    def add_arguments(parser):
        add = parser.add_argument

        add(
            '--model-name',
            '-m',
            type=str,
            help="OpenAI model name. model_name must be one of [gpt-4, gpt-3.5-turbo, gpt-3.5-turbo-16k]. (default: gpt-4)",
            choices=['gpt-4', 'gpt-3.5-turbo', 'gpt-3.5-turbo-16k'],
            default='gpt-4'
        )
        add(
            '--temperature',
            '-t',
            help='Temperature of LLM models. The lower the temperature, the more accurate the answer, and the higher the temperature, the more variable the answer. (default: 0.1)',
            type=float,
            default=0.1,
        )

        add(
            '--accelerator',
            '-a',
            help='Device name for MOFTransformer. accelerator must be one of [cuda, gpu, cpu]',
            type=str,
            default='cuda'
        )
        add(
            '--max-length-in-predictor',
            type=int,
            default=30,
        )
        add(
            '--num-genetic-cycle',
            type=int,
            default=3,
        )
        add(
            '--logger',
            help='logger for generation task (Default: generate_mof.log)',
            type=str,
            default='generate_mof.log'
        )


    @staticmethod
    def run(args):
        from chatmof.main import main
        kwargs = {}
        print (dict(args))
        print (args)
        main(kwargs)


import re
from typing import List
from pathlib import Path


def preprocess_json_input(input_str: str) -> str:
    # Replace single backslashes with double backslashes,
    # while leaving already escaped ones intact
    corrected_str = re.sub(
        r'(?<!\\)\\(?!["\\/bfnrt]|u[0-9a-fA-F]{4})', r"\\\\", input_str
    )
    return corrected_str


def search_file(name: str, direc: Path) -> List[Path]:
    name = name.strip()
    if '*' in name:
        return list(direc.glob(name))
    
    f_name = direc/name
    if f_name.exists():
        return [f_name]
    
    return False


from langchain.chat_models import ChatOpenAI
from chatmof.agents.agent import ChatMOF
from langchain.callbacks import StdOutCallbackHandler

import io
import re
from contextlib import redirect_stdout
import pandas as pd
import random
from pathlib import Path

files = Path('/storage/dudgns1675/data/moftransformer/coremof/4_descriptor/total')
names = [cif.stem for cif in files.glob('*.grid')]


with open('test/generator.txt') as f:
    questions = f.readlines()

verbose = True
search_internet = False

model='gpt-4'
llm = ChatOpenAI(temperature=0, model=model)
callback_manager = [StdOutCallbackHandler()]

chatmof = ChatMOF.from_llm(
    llm=llm, 
    verbose=verbose, 
)

qna = []

save_path = Path(f'./test/generator_{model}')
save_path.mkdir(exist_ok=True, parents=True)
for i, question in enumerate(questions):
    if i in [0, 1,  2, 4, 5, 7, 8]:
        continue
    
    question = question.strip()

    with io.StringIO() as buf, redirect_stdout(buf):
        try:
            output = chatmof.run(question, callbacks=callback_manager)
        except Exception as e:
            print (type(e), e)
            output = None
        log = buf.getvalue()

    qna.append([i, question, output])

    with (save_path/f'{i}.output').open('w') as f:
        f.write(log)

df = pd.DataFrame(qna)
df.to_csv(f'test/generator_{model}.csv')


import re
import json
import pandas as pd
from pathlib import Path
from typing import Dict, Any, List, Optional
from langchain.base_language import BaseLanguageModel
from langchain.chains.base import Chain
from langchain.chains.llm import LLMChain
from langchain.prompts import BasePromptTemplate, PromptTemplate
from langchain.callbacks.manager import CallbackManagerForChainRun
from langchain.agents import create_pandas_dataframe_agent

from chatmof import __root_dir__
from chatmof.config import config
from chatmof.tools.predictor.runner import MOFTransformerRunner
from chatmof.tools.predictor.utils import _predictable_properties, model_names
from chatmof.tools.search_csv.base import TableSearcher
from chatmof.tools.genetic_algorithm.genetic_algorithm import GeneticAlgorithmChain
from chatmof.tools.genetic_algorithm.prompt import PLAN_PROMPT
from chatmof.tools.genetic_algorithm.cif_generate import CIFGenerator


class Generator(Chain):
    llm: BaseLanguageModel
    llm_chain: LLMChain
    generator_chain: GeneticAlgorithmChain
    topologies: List[str] = config['topologies']
    input_key: str = 'question'
    output_key: str = 'answer'

    @property
    def input_keys(self) -> List[str]:
        return [self.input_key]
    
    @property
    def output_keys(self) -> List[str]:
        return [self.output_key]
    
    def _write_log(self, action, text, run_manager):
        run_manager.on_text(f"\n[Generator] {action}: ", verbose=self.verbose)
        run_manager.on_text(text, verbose=self.verbose, color='yellow')
    
    def _call(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ):
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        callbacks = _run_manager.get_child()

        llm_output = self.llm_chain.run(
           question=inputs[self.input_key],
           model_names=model_names,
           stop=['Question:'],
           callbacks=callbacks
        )

        output = self._parse_output(llm_output)
        
        self._write_log('Thought', output['Thought'], run_manager)
        self._write_log('Objective', output['Objective'], run_manager)
        
        df_dict = dict()
        prop = output['Property']
        self._write_log('Predict Properties', prop+"\n", run_manager)
        for topology in self.topologies:
            df, info_ls = self.run_predictor(prop_text=prop, topology=topology, data_dir=config['hmof_dir'])
            df_dict[topology] = df

        save_path = Path(config['generate_dir'])/f'{prop}-0.csv'
        pd.concat(df_dict.values()).to_csv(str(save_path))

        for cycle in range(1, config['num_genetic_cycle']+1):
            run_manager.on_text(f'\n\n[Generator] Run genetic algorithm : Cycle {cycle}.\n', color='green')
            df_dict = self.run_genetic(df_dict, output, run_manager, cycle, info_ls)

        self._write_log('Final Thought', output['Final Thought'], run_manager)
        
        df_final = df_dict[self.topologies[0]]
        for topo in self.topologies[1:]:
            df_gen = df_dict[topo]
            df_final = df_final.merge(df_gen, how='outer')

        information = f'Information of models : {info_ls}. If unit or condition are existed, you must include it in the final output.'
        searcher = TableSearcher.from_dataframe(
            llm = self.llm,
            dataframe = df_final,
            verbose=self.verbose,
            run_manager=run_manager,
        )

        final_output = searcher.run(
            question=output['Final Thought'],
            run_manager=run_manager,
            information=information,
        )
        return {self.output_key: final_output}


    def run_genetic(self, df_dict, output, run_manager, cycle, info_ls):
        self._write_log('Find Parents', output['Search'], run_manager)

        prop = output['Property']
        direc = Path(config['generate_dir']) / f'{prop}-{cycle}'

        parent_dict = dict()
        information = f'Information of models : {info_ls}. If unit or condition are existed, you must include it in the final output.'

        for topology in self.topologies:
            searcher = TableSearcher.from_dataframe(
                llm=self.llm,
                dataframe = df_dict[topology],
                verbose=self.verbose,
                run_manager=run_manager,
            )            
            prompt = "{} (Objective: {}). If you need to set a range of data, you need to make it wide to make the data exist."\
                    .format(output['Search'], output['Objective'])
            search_output = searcher.run(
                question=prompt, 
                return_observation=True,
                run_manager = run_manager,
                information=information,
            )
            parents = self._parse_predictor(search_output)

            if not parents:
                continue
                
            parent_dict[topology] = parents

        if not parent_dict:
            raise ValueError('There are no parents') 
    
        self._write_log('Get Children', output['Generate'], run_manager)

        child_dict = dict()
        for topology in self.topologies:
            if not topology in parent_dict.keys():
                continue

            children = self.generator_chain.run(
                question=output['Generate'],
                parents=parent_dict[topology],
                run_manager=run_manager,
            )
            child_dict[topology] = children
            print (children)

        self._write_log('Generate Structures', '', run_manager)

        generator = CIFGenerator(direc)
        for topology in self.topologies:
            if topology not in child_dict.keys():
                continue
            generator.run(topology=topology, cif_list=child_dict[topology])

        self._write_log('Predict Properties', output['Property']+"\n", run_manager)
        
        df_ls = []
        for topology in self.topologies:
            if topology not in child_dict.keys():
                continue
            try:
                df_gen, info_ls = self.run_predictor(
                    prop_text=output['Property'], 
                    topology=topology, 
                    data_dir=direc,
                )
            except ValueError:
                continue

            df_dict[topology] = df_dict[topology].merge(df_gen, how='outer')
            df_ls.append(df_gen)

        pd.concat(df_ls).to_csv(str(direc/f'../{prop}-{cycle}.csv'))
        return df_dict

    def _parse_output(self, text):
        thought = re.search(r"Thought:\s*(.+?)\s*\n", text, re.DOTALL)
        search = re.search(r"Search look-up table:\s*(.+?)\s*\n", text, re.DOTALL)
        prop = re.search(r"Property:\s*(.+?)\s*\n", text, re.DOTALL)
        objective = re.search(r"Objective:\s*(.+?)\s*\n", text, re.DOTALL)
        genetic_algorithm = re.search(r"Genetic algorithm:\s*(.+?)\s*(\n|$)", text, re.DOTALL)
        final_thought = re.search(r"Final Thought:\s*(.+?)\s*(\n|$)", text, re.DOTALL)

        if not thought:
            raise ValueError(f'unknown format from LLM: {text}')
        if not search:
            raise ValueError(f'unknown format from LLM: {text}')
        if not objective:
            raise ValueError(f'unknown format from LLM: {text}')
        if not genetic_algorithm:
            raise ValueError(f'unknown format from LLM: {text}')
        if not prop:
            raise ValueError(f'unknown format from LLM: {text}')
        if not final_thought:
            raise ValueError(f'unknown format from LLM: {text}')
        
        return {
            'Thought': thought.group(1),
            'Property': prop.group(1),
            'Objective': objective.group(1),
            'Search': search.group(1),
            'Generate': genetic_algorithm.group(1),
            'Final Thought': final_thought.group(1),
        }

    def _parse_property(self, prop_text):
        prop_ls = re.split(',\s*', prop_text)
        for prop in prop_ls:
            if prop not in _predictable_properties:
                raise ValueError(f'property should be one of [{model_names}], not {prop}')
        return prop_ls
    
    def run_predictor(self, prop_text: str, topology: str, data_dir: str) -> List[str]:
        runner = MOFTransformerRunner(
            model_dir=config['model_dir'],
            data_dir=str(data_dir),
            verbose=self.verbose,
        )
        
        material = f'{topology}*.cif'
        df_ls = []
        info_ls = []
        for prop in self._parse_property(prop_text):
            cif_id, logits, model_info = runner.run(prop, material)
            df = pd.DataFrame({'cif_id': cif_id, prop: logits})
            df_ls.append(df)
            info_ls.append(model_info)

        df_total = df_ls[0]
        for df in df_ls[1:]:
            df_total.merge(df, on='cif_id', how='outer')

        return df, model_info

    def _parse_predictor(self, output: str) -> List[List[str]]: 
        try:
            data_ls = json.loads(output)
            if isinstance(data_ls, list):
                return data_ls
            elif isinstance(data_ls, dict):
                if 'cif_id' in data_ls:
                    return zip(*data_ls.values())
                else:
                    return data_ls.items()
            else:
                raise ValueError()
        except json.JSONDecodeError:
            pass

        label, parents = self._parse_markdown(output)
        return parents        
            
    def _parse_markdown(self, output: str) -> List[List[str]]:
        parents = []
        lines = [line.replace("|", "").strip() for line in output.split('\n')]
        label = lines[0].split()

        for line in lines[1:]:
            data = re.split(r',\s*|\s+', line)
            if len(data) < 2:
                continue
            elif "+" in data[0]:
                parents.append(data)
            elif "+" in data[1]:
                parents.append(data[1:])
            else:
                raise ValueError(output)
        return label, parents

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        prompt: BasePromptTemplate = PLAN_PROMPT,
        **kwargs: Any
    ) -> Chain:
        prompt = PromptTemplate(
            template=prompt, 
            partial_variables={'model_names': model_names},
            input_variables=['question']
        )
        llm_chain = LLMChain(llm=llm, prompt=prompt)
        genetic_chain = GeneticAlgorithmChain.from_llm(llm)

        return cls(
            llm=llm,
            llm_chain=llm_chain, 
            generator_chain=genetic_chain,
            **kwargs)


if __name__ == '__main__':
    from langchain.chat_models import ChatOpenAI
    from matplotlib import pyplot as plt
    
    llm = ChatOpenAI(temperature=0)
    #chain = GeneticAlgorithmChain.from_llm(llm, verbose=True)

    prompt = 'generate structures with bandgap near 0.5.'
    #prompt = 'generate structures with highest bandgap'
    #prompt = 'create a new material with a surface area close to 0.9 from 200 materials'
    chain = Generator.from_llm(llm, verbose=True)
    output = chain({'question':prompt})
    print (output['answer'])

    df = output['origin_df']
    df2 = output['generate_df']
    
    plt.hist(df['bandgap'], density=True, alpha=0.7, color='royalblue')
    plt.hist(df2['bandgap'], density=True, alpha=0.7, color='salmon')
    plt.legend(['original', 'generate'])
    plt.xlabel('Band-gap (eV)', fontsize=15)
    plt.ylabel('Density', fontsize=15)
    plt.show()


    


import os
from typing import List, Optional
from pathlib import Path
from tqdm import tqdm
import logging
from timeout_decorator import timeout

import pormake as pm
from moftransformer.utils.prepare_data import make_prepared_data
from chatmof.config import config


pm.log.disable_print()
pm.log.disable_file_print()   

DATABASE = pm.Database()
builder = pm.Builder()


def get_logger(filename):
    if os.path.exists(filename):
        os.unlink(filename)

    logger = logging.getLogger(filename)
    logger.setLevel(logging.ERROR)

    formatter = logging.Formatter(
        fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    file_handler = logging.FileHandler(filename)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    return logger

logger = get_logger(config['logger'])


class CIFGenerator(object):
    def __init__(self, save_dir:str) -> None:
        self.save_dir = Path(save_dir)
        self.save_dir.mkdir(exist_ok=True, parents=True)
        
    def run(
        self, 
        topology: str,
        cif_list: Optional[List[str]] = None,
    ) -> None:
        topo = DATABASE.get_topo(topology)

        for cif in tqdm(cif_list, desc=topology):
            save_path = self.save_dir/f'{topology}+{cif}.cif'
            success = self._run_cif(cif, topo, save_path)                      

    def _run_cif(
            self,
            cif: str,
            topo: pm.Topology,
            save_path: Path,
    ):
        try:
            if not save_path.exists():
                current_cif = self._generate_cif(cif, topo)
                current_cif.write_cif(str(save_path))

            is_success = make_prepared_data(
                save_path, 
                self.save_dir, 
                logger=logger, 
                eg_logger=logger
            )

            if not is_success:
                save_path.unlink()
                return False
            else:
                return True
            
        except Exception as e:
            if save_path.exists():
                save_path.unlink()
            logger.error(e)
            return False

    @timeout(30) # timeout set to 30 second
    def _generate_cif(
        self,
        cif: str,
        topo: pm.Topology,
    ):
        bb_name1, bb_name2 = cif.split('+')
        bb1 = DATABASE.get_bb(bb_name1)
        bb2 = DATABASE.get_bb(bb_name2)

        # No metal cluster or more than 2 metal cluster
        if (bb1.has_metal and bb2.has_metal) or (not bb1.has_metal and not bb2.has_metal):
            return None   
        
        current_nodes = {}

        if bb_name2.startswith('E'):
            current_nodes[0] = bb1
            current_edges = {
                tuple(et): bb2 for et in topo.unique_edge_types
            }
        else:
            current_nodes[0] = bb1
            current_nodes[1] = bb2
            current_edges = {}

        current_mof = builder.build_by_type(topo, current_nodes, current_edges)
        return current_mof







import os
from typing import Any
from langchain.tools import Tool
from langchain.base_language import BaseLanguageModel
from langchain.tools.base import BaseTool

from chatmof.config import config
from chatmof.tools.search_csv.base import TableSearcher


def _get_search_csv(
        llm: BaseLanguageModel,
        file_path: str = config['lookup_dir'],
        verbose: bool = False,
        **kwargs: Any) -> BaseTool:

    return Tool(
        name="search_csv",
        description=(
                "A tools that extract accurate properties from a look-up table in the database. "
                #"input must be provided in the form of a full sentence. "
                "The input must be a detailed full sentence to answer the question."
        ),
        func=TableSearcher.from_filepath(
            llm=llm, file_path=file_path, verbose=verbose
        ).run
    )



PROMPT = """plan to use machine learning to predict the properties of matter. To answer the question, you have to fill in the following format:

Question: the input question you must answer
Thought: you should always think about what to do
Property: the property you can predict, should be one of [{model_names}]
Material: names of materials separated using comma. If you need to proceed for all material, write *. To proceed for a specific topology, append the topology name with an * (ex. pcu*)
... (this Property/Material can repeat N times)
Run Machine Learning: nothing to do
Final Thought: you should think about how you will derive a final answer from the results of machine learning.

Begin!

Question: predict the surface area and hydrogen uptake of ACOPEN and ELOBPE.
Thought: I need to gather data on ACOPEN and train a machine learning model to predict its surface area and hydrogen uptake.
Property: surface_area
Material: ACOPEN, ELOBPE
Property: hydrogen_uptake_100bar_77K
Material: ACOPEN, ELOBPE
Final Thought: Based on the result, answer the question using predicted surface area and the predicted hydrogen uptake at 100 bar.

Question: which MOF has a highest band-gap?
Thought: I need to gather data on the band-gap of different structures and compare them to find the one with the highest value.
Property: bandgap
Material: *
Final Thought: Based on the result, find the structure with the highest predicted band-gap value.

Question: Predict surface area and save results in csv format, only pcu topology.
Thought: I need to gather data on the surface area of materials with pcu topology and train a machine learning model to predict their surface area. Then, I can use the model to make predictions and save the results in a csv format.
Property: surface_area
Material: pcu*
Final Thought: The predicted surface area values for materials with pcu topology can be found in the saved csv file.

Question: {question}"""


FINAL_MARKDOWN_PROPMT = """
You need to answer the question from the markdown table below

Markdown Table:
{table}

{information}

Question: {question}
Answer:"""


import os
from pathlib import Path
import re
import json
from typing import List
from pathlib import Path
from pydantic import BaseModel
from chatmof import __root_dir__
from chatmof.config import config
from chatmof.tools.predictor.utils import predict, model_names, search_file


class MOFTransformerRunner(BaseModel):
    model_dir: str = config['model_dir']
    data_dir: str = config['data_dir']
    verbose: bool = False

    def run(self, 
            prop: str, 
            material: str):
        if prop not in model_names:
            raise ValueError(f'property should be one {model_names}, not {property}')

        #params = os.path.join(self.model_dir, f'{prop}/hparams.yaml')
        model_dir = Path(self.model_dir)/prop
        data_list = self.parse_data(material)

        output = predict(
            data_list=data_list,
            model_dir=model_dir,
            verbose=self.verbose,
        )

        with (model_dir/'model_info.json').open() as f:
            model_info = json.load(f)

        if 'regression_logits' in output:
            cif_id = output['cif_id']
            logits = output['regression_logits']
        elif 'classification_logits_index' in output:
            label_json = model_dir/'label.json'
            if not label_json.exists():
                raise FileNotFoundError(f'There are no "label.json" in {model_dir}')
            with label_json.open() as f:
                labels = json.load(f)
            
            cif_id = output['cif_id']
            logits = [labels[i] for i in output['classification_logits_index']]
            
        return cif_id, logits, model_info
        
    def parse_data(self, material:str) -> List[Path]:
        data_dir = Path(self.data_dir)
        data_list = []

        s_mat = re.split(r",\s*", material)
        for mat in s_mat:
            if not mat.endswith('.cif'):
                if mat.endswith('*'):
                    mat = f'{mat}.cif'
                else:
                    mat = f'{mat}*.cif'

            if f_mat := search_file(mat, data_dir):
                data_list.extend(f_mat)

        if len(data_list) < len(s_mat):
            raise ValueError(f'There are no data in {material}')
        
        return data_list
    

if __name__ == '__main__':
    from chatmof.config import config

    runner = MOFTransformerRunner(
        model_dir=config['model_dir'], 
        data_dir=config['data_dir']
    )
    output = runner.run('solvent_removal_stability', 'PUG*')
    print (output)


import os
import re
from pathlib import Path
import pandas as pd
from typing import Dict, Any, List, Optional

from langchain.base_language import BaseLanguageModel
from langchain.chains.base import Chain
from langchain.chains.llm import LLMChain
from langchain.prompts import PromptTemplate
from langchain.callbacks.manager import CallbackManagerForChainRun
from langchain.tools.python.tool import PythonAstREPLTool
import tiktoken

from chatmof.config import config
from chatmof.tools.search_csv.prompt import DF_PROMPT


class TableSearcher(Chain):
    """Tools that search csv using Pandas agent"""
    llm_chain: LLMChain
    df: pd.DataFrame
    encoder: tiktoken.core.Encoding
    num_max_data: int = 200
    input_key: str = 'question'
    output_key: str = 'answer'

    @property
    def input_keys(self) -> List[str]:
        return [self.input_key]
    
    @property
    def output_keys(self) -> List[str]:
        return [self.output_key]
    
    def _parse_output(self, text:str) -> Dict[str, Any]:
        thought = re.search(r"(?<!Final )Thought:\s*(.+?)\s*(Observation|Final|Input|Thought|Question|$)", text, re.DOTALL)
        input_ = re.search(r"Input:\s*(?:```|`)?(.+?)(?:```|`)?\s*(Observation|Final|Input|Thought|Question|$)", text, re.DOTALL)
        final_thought = re.search(r"Final Thought:\s*(.+?)\s*(Observation|Final|Input|Thought|Question|$)", text, re.DOTALL)
        final_answer = re.search(r"Final Answer:\s*(.+?)\s*(Observation|Final|Input|Thought|Question|$)", text, re.DOTALL)
        observation = re.search(r"Observation:\s*(.+?)\s*(Observation|Final|Input|Thought|Question|$)", text, re.DOTALL)
        
        if (not input_) and (not final_answer):
            raise ValueError(f'unknown format from LLM: {text}')
        
        return {
            'Thought': (thought.group(1) if thought else None),
            'Input': (input_.group(1).strip() if input_ else None),
            'Final Thought' : (final_thought.group(1) if final_thought else None),
            'Final Answer': (final_answer.group(1) if final_answer else None),
            'Observation': (observation.group(1) if observation else None),
        }
    
    def _clear_name(self, text:str) -> str:
        remove_list = ['_clean_h', '_clean', '_charged', '_manual', '_ion_b', '_auto', '_SL', ]
        str_remove_list = r"|".join(remove_list)
        return re.sub(rf"({str_remove_list})", "", text)
    
    @staticmethod
    def _get_df(file_path: str):
        if isinstance(file_path, str):
            file_path = Path(file_path)
            
        if file_path.suffix == '.csv':
            df = pd.read_csv(file_path)
        elif file_path.suffix == '.xlsx':
            df = pd.read_excel(file_path)
        elif file_path.suffix == '.json':
            df = pd.read_json(file_path)
        else:
            raise ValueError(f'table must be .csv, .xlsx, or .json, not {file_path.suffix}')

        return df
    
    def _write_log(self, action, text, run_manager):
        run_manager.on_text(f"\n[Table Searcher] {action}: ", verbose=self.verbose)
        run_manager.on_text(text, verbose=self.verbose, color='yellow')
    
    def _call(
            self,
            inputs: Dict[str, Any],
            run_manager: Optional[CallbackManagerForChainRun] = None
    ):
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        callbacks = _run_manager.get_child()
        
        return_observation = inputs.get('return_observation', False)
        information = inputs.get('information', "If unit exists, you must include it in the final output. The name of the material exists in the column \"name\".")

        agent_scratchpad = ''
        max_iteration = config['max_iteration']

        input_ = self._clear_name(inputs[self.input_key])

        for i in range(max_iteration + 1):
            llm_output = self.llm_chain.run(
                df_index = str(list(self.df)),
                information = information,
                df_head = self.df.head().to_markdown(),
                question=input_,
                agent_scratchpad = agent_scratchpad,
                callbacks=callbacks,
                stop=['Observation:', 'Question:',]
            )

            if not llm_output.strip():
                agent_scratchpad += 'Thought: '
                llm_output = self.llm_chain.run(
                    df_index = str(list(self.df)),
                    information = information,
                    df_head = self.df.head().to_markdown(),
                    question=input_,
                    agent_scratchpad = agent_scratchpad,
                    callbacks=callbacks,
                    stop=['Observation:', 'Question:',]
                )
            
            #if llm_output.endswith('Final Answer: success'):
            if re.search(r'Final Answer: (success|.* above|.* DataFrames?).?$', llm_output):
                thought = f'Final Thought: we have to answer the question `{input_}` using observation\n'
                agent_scratchpad += thought
                llm_output = self.llm_chain.run(
                    df_index = str(list(self.df)),
                    df_head = self.df.head().to_markdown(),
                    question=input_,
                    information = information,
                    agent_scratchpad = agent_scratchpad,
                    callbacks=callbacks,
                    stop=['Observation:', 'Question:',]
                )
                llm_output = thought + llm_output

            output = self._parse_output(llm_output)

            if output['Final Answer']:
                if output['Observation']:
                    raise ValueError(llm_output)
                
                self._write_log('Final Thought', output['Final Thought'], run_manager)

                final_answer: str = output['Final Answer']
                #if final_answer == 'success':
                #    agent_scratchpad += 'Thought: {}\nInput: {} \nObservation: {}\n'\
                #            .format(output['Thought'], output['Input'], observation)

                #check_sentence = ' Check to see if this answer can be your final answer, and if so, you should submit your final answer. You should not do any additional verification on the answer.'
                check_sentence = ''
                if re.search(r'nothing', final_answer):
                    final_answer = 'There are no data in database.' # please use tool `predictor` to get answer.'
                elif final_answer.endswith('.'):    
                    final_answer += check_sentence
                else:
                    final_answer = f'The answer for question "{input_}" is {final_answer}.{check_sentence}'
                return {self.output_key: final_answer}
            
            elif i >= max_iteration:
                final_answer = 'There are no data in database'
                self._write_log('Final Thought', final_answer, run_manager)
                return {self.output_key: final_answer}

            else:
                self._write_log('Thought', output['Thought'], run_manager)
                self._write_log('Input', output['Input'], run_manager)
                
            pytool = PythonAstREPLTool(locals={'df':self.df})
            observation = str(pytool.run(output['Input'])).strip()

            num_tokens = len(self.encoder.encode(observation))
            
            if return_observation:
                if "\n" in observation:
                    self._write_log('Observation', "\n"+observation, run_manager)
                else:
                    self._write_log('Observation', observation, run_manager)
                return {self.output_key: observation}
            
            if num_tokens > 3400:
                raise ValueError('The number of tokens has been exceeded.')
                observation = f"The number of tokens has been exceeded. To reduce the length of the message, please modify code to only pull up to {self.num_max_data} data."
                self.num_max_data = self.num_max_data // 2

            if "\n" in observation:
                self._write_log('Observation', "\n"+observation, run_manager)
            else:
                self._write_log('Observation', observation, run_manager)

            agent_scratchpad += 'Thought: {}\nInput: {} \nObservation: {}\n'\
                .format(output['Thought'], output['Input'], observation)

        raise AssertionError('Code Error! please report to author!')

    @classmethod
    def from_filepath(
        cls,
        llm: BaseLanguageModel,
        file_path: Path = Path(config['lookup_dir']),
        prompt: str = DF_PROMPT,
        **kwargs
    ) -> Chain:
        template = PromptTemplate(
            template=prompt,
            input_variables=['df_index', 'information', 'df_head', 'question', 'agent_scratchpad']
        )
        llm_chain = LLMChain(llm=llm, prompt=template)
        df = cls._get_df(file_path)
        encoder = tiktoken.encoding_for_model(llm.model_name)
        return cls(llm_chain=llm_chain, df=df, encoder=encoder, **kwargs)
    
    @classmethod
    def from_dataframe(
        cls,
        llm: BaseLanguageModel,
        dataframe: pd.DataFrame,
        prompt: str = DF_PROMPT,
        **kwargs
    ) -> Chain:
        template = PromptTemplate(
            template=prompt,
            input_variables=['df_index', 'information', 'df_head', 'question', 'agent_scratchpad']
        )
        llm_chain = LLMChain(llm=llm, prompt=template)
        encoder = tiktoken.encoding_for_model(llm.model_name)
        return cls(llm_chain=llm_chain, df=dataframe, encoder=encoder, **kwargs)


from typing import Any, Dict
from pathlib import Path
from pydantic import BaseModel, root_validator
from langchain.tools import Tool
from langchain.base_language import BaseLanguageModel
from langchain.tools.base import BaseTool
from chatmof import __root_dir__
from chatmof.tools.predictor.base import Predictor



def _get_predict_properties(
        llm: BaseLanguageModel,
        verbose: bool = False,
        **kwargs: Any
) -> BaseTool:
    return Tool(
        name="predictor",
        description=(
            "Useful tool to predict material properties using machine learning. "
            "More imprecise than the search_csv tool, but can be used universally. "
            "You must use predictor before using google_search or wikipedia. "
            "The input must be a detailed full sentence to answer the question."
        ),
        func=Predictor.from_llm(llm=llm, verbose=verbose).run,
    )


PLAN_PROMPT = """Create a plan to generate material based on the following question. 
Use the following format:

Question: the input question you must to answer
Thought: you should always think about what to do
Property: the property you can predict, must be one of [{model_names}]
Objective: you should decide what criteria you wan to generate by.
Search look-up table: plan to extract 100 material for the purpose from the look-up table where the property is pre-calculated.
Genetic algorithm: plan to create a new materials using the 100 extracted materials.
Final thought: get a final answer based on the structures you generate.

Begin!
Question: generate a material with a porosity of 0.5 and surface area of 120 m2/g
Thought: I need to generate a material with a porosity value of 0.5 and surface area of 120 m2/g. 
Property: void_fraction, accessible_surface_area
Objective: near 0.5, near 120
Search look-up table: extract cif_id, void_fraction, and accessible_surface_area of 100 materials with porosity close to 0.5 and surface area near 120 m2/g from look-up tables. 
Genetic algorithm: create a new material with a porosity close to 0.5 and surface area near 120 m2/g from 100 materials
Final Thought: Based on the generated CIF, find the material that is closest to a porosity of 0.5 and a surface area of 120 m2/g.

Question: generate a material with a highest band-gap
Thought: I need to generate a material with a highest band-gap.
Property: bandgap
Objective: max
Search look-up table: extract cif_id and bandgap of 100 materials with high band-gap value from look-up tables. 
Genetic algorithm: generate 100 new materials with the highest band gap from the 100 materials.
Final Thought: Based on the generated CIF, find the material that has highest band-gap.

Question: {question}"""


GENETIC_PROMPT = (
    "You should act as a generator to find the optimal material. "
    "A substance consists of a block1, block2, and must maintain the order. "
    "I will give you 100 parent materials. "
    "Based on these, you must answer as many new children as you expect to answer the question. "
    "The block1 and block2 used in the child must be the blocks used in the parent, and you must not create blocks that don't exist. "
    "You must generate children diverse. "
    "The children must not be duplicates of existing parents or already created children. "
    "You output children only and nothing else."
    "\n\n"
    "Begin.\n"
    "Question: {question}\n"
    "Parent:\n"
    "V12+T31, V24+T32, V7+T12\n"
    "4 new Children:\n"
    "V12+T12, V24+T31, V7+T31, V7+T32\n\n"
    "Parent:\n"
    "{parents}"
    "100 new Children:\n"
)


import os
from chatmof import __root_dir__


config = {
    # ChatMOF
    'search_internet': False,
    'verbose': True,

    # LLM - openAI
    'temperature': 0.1,    
    'model': 'gpt-4',

    # data directory
    'model_dir': os.path.join(__root_dir__, 'database/load_model/'),
    'structure_dir': os.path.join(__root_dir__, 'database/structures/'),
    'data_dir': os.path.join(__root_dir__, 'database/structures/coremof/'),
    'hmof_dir': os.path.join(__root_dir__, 'database/structures/hMOF/'),
    'generate_dir': os.path.join(__root_dir__, 'database/structures/generate'),

    # table searcher
    'lookup_dir': os.path.join(__root_dir__, 'database/tables/coremof.xlsx'),
    'max_iteration': 3,

    # building block searcher
    'buildingblock_dir' : os.path.join(__root_dir__, 'database/tables/mofkey.xlsx'),

    # predictor
    'max_length_in_predictor' : 30,
    'accelerator' : 'cuda',

    # generator
    'num_genetic_cycle': 3,
    'num_parents': 200,
    'logger': 'generate_mof.log',
    'topologies': ['pcu', 'dia', 'acs', 'rtl', 'cds', 'srs', 'ths', 'bcu', 'fsc'],
}


from langchain.llms import OpenAI
from langchain.chat_models import ChatOpenAI
from langchain.callbacks import StdOutCallbackHandler
from chatmof.agents.agent import ChatMOF
from chatmof.config import config


if __name__ == '__main__':
    #question = "What is the surface area of ACOGEF?"
    #question = "Tell me structures that has void fraction between 0.2 and 0.3"
    #question = "What is top-3 structures that have high PLD?"
    #question = "Can you tell me the type of open metal in DUQYEU01?"
    #question = "Tell me top-5 structures that has void fraction near 0.7."
    #question = 'Generate the structures with highest surface area and visualize it.'
    #question = "Does UMODEH02_clean has open metal site or not?"
    #question = "What's the largest cavity diameter in materials with a density greater than a 7.0?"
    #question = "generate MOF with volume fraction == 0.5"
    #question = "What is the hydrogen uptake of ACOGEF_clean?"
    #question = "How does the non-accessible surface area of QEZZEC compare with other materials?"
    #question = "What is the pore volume and surface area of ACOGEF_clean?"
    #question = "Can you predict all bandgap prediction in coremof?"
    #question = "What is the highest bandgap MOF?"
    #question = "What is the bandgap of ACOGEF_clean and ABETAE_clean?"
    #question = "What is the void fraction of ACOGEF_clean and ABETAE_clean?"
    #question = "Can you check that ACOGEF is in directory /home/dudgns1675/autogpt/ChatMOF/chatmof/database/structures/raw ?"
    #question = "Can you provide the top 10 materials with the highest largest free pore diameter?"
    #question = 'What is the surface area of MIBQAR16?'
    #question = "Can VAHSON have an open metal site?"
    question = "How does the pore limiting diameter of YUSGID_clean compare with other materials?"
    #question = "At room temperature (298K), what's the CO2 Henry coefficient for XEGKUR_clean?"

    verbose = True
    search_internet = False
    model = 'gpt-4'

    llm = ChatOpenAI(temperature=config['temperature'], model=model)
    callback_manager = [StdOutCallbackHandler()]

    chatmof = ChatMOF.from_llm(
        llm=llm, 
        verbose=verbose, 
        search_internet=search_internet,
    )
    output = chatmof.run(question, callbacks=callback_manager)



PROMPT = """plan to visualize the structure. To answer the question, you have to fill in the following format:

Question: the input question you must answer
Thought: you should always think about what to do
Material: names of materials separated using comma. If you need to proceed for all material, write *. To proceed for a specific topology, append the topology name with an * (ex. pcu*)

Begin!

Question: Visualize the ZVBEFJ and acs+N123+E132.
Thought: I need to visualize the structure ZVBEFJ and acs+N123+E132.
Material: ZVBEFJ, acs+N123+E132

Question: Visualize the all mofs with pcu topology
Thought: I need to visualize the structures with pcu topology.
Material: pcu*

Question: {question}"""


"""A tool for running python code in a REPL."""

import ast
import asyncio
import re
import sys
from contextlib import redirect_stdout
from io import StringIO
from typing import Any, Dict, Optional

from pydantic import Field, root_validator

from langchain.callbacks.manager import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)
from langchain.tools.base import BaseTool
from langchain.utilities import PythonREPL


def _get_default_python_repl() -> PythonREPL:
    return PythonREPL(_globals=globals(), _locals=None)


def sanitize_input(query: str) -> str:
    """Sanitize input to the python REPL.
    Remove whitespace, backtick & python (if llm mistakes python console as terminal)

    Args:
        query: The query to sanitize

    Returns:
        str: The sanitized query
    """

    # Removes `, whitespace & python from start
    query = re.sub(r"^(\s|`)*(?i:python)?\s*", "", query)
    # Removes whitespace & ` from end
    query = re.sub(r"(\s|`)*$", "", query)
    return query


class PythonREPLTool(BaseTool):
    """A tool for running python code in a REPL."""

    name = "Python_REPL"
    description = (
        "A Python shell. Use this to execute python commands. "
        "Input should be a valid python command. "
        "If you want to see the output of a value, you should print it out "
        "with `print(...)`."
    )
    python_repl: PythonREPL = Field(_get_default_python_repl)
    sanitize_input: bool = True

    def _cleanup(self, query):
        query = query.replace("\\n", "\n")
        return query

    def _run(
        self,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> Any:
        """Use the tool."""
        if self.sanitize_input:
            query = sanitize_input(query)
        query = self._cleanup(query)
        return self.python_repl.default().run(query)

    async def _arun(
        self,
        query: str,
        run_manager: Optional[AsyncCallbackManagerForToolRun] = None,
    ) -> Any:
        """Use the tool asynchronously."""
        if self.sanitize_input:
            query = sanitize_input(query)
        query = self._cleanup(query)
        loop = asyncio.get_running_loop()
        result = await loop.run_in_executor(None, self.run, query)

        return result


class CLICommand:
    """
    Inital setup for ChatMOF. Download coremof, hmof, and load_models
    """

    @staticmethod
    def add_arguments(parser):
        pass

    @staticmethod
    def run(args):
        from chatmof.setup_module import setup
        
        setup()